-- For I18N support to mark translatable strings, plus a minimal formatter
require 'strict'

local mt = {}

-- fmt is in Java MessageFormat style, like:
-- "Storage unit {0} out of space at {1} bytes"
-- See:
-- http://docs.oracle.com/javase/7/docs/api/java/text/MessageFormat.html
function mt:format(fmt, ...)
    local msg = type(fmt) == 'table' and fmt or {fmt, ...}
    assert(type(msg[1]) == 'string')
    -- KLUDGE doesn't yet handle forms like {0,date} or more complex variants
    local formatted = msg[1]:gsub('{(%d+)}', function(n)
        return tostring(msg[tonumber(n) + 2])
    end)
    return formatted
end

function mt:__(str)
    return str
end

local m = {}

setmetatable(m, {
    __index = mt,
    __call = function(t, str)
        return t:__(str)
    end,
})

return m
