local config = require 'system_config'
local u = require 'unix'

local exrfd
local exrmap

local get_fpga_version = {
    edaqxr = function()
        if not exrfd then
            exrfd = u.open('/dev/mem', u.O_RDWR)
            if not exrfd then return "Unknown" end
            exrmap = u.mmap(0x100, exrfd, 0xff200000)
            u.close(exrfd)
            if not exrmap then return "Unknown" end
        end
        if not exrmap then return "Unknown" end
        local version_major = exrmap:r32(0x90)
        local version_minor = exrmap:r32(0x94)
        local version_sub_minor = exrmap:r32(0x98)
        local version_extra = exrmap:r32(0x9c)
        return string.format("%d.%d.%d.%d", version_major, version_minor,
            version_sub_minor, version_extra), true
    end,
    cx23r = function()
        local fh = io.open('/var/log/fpga-hardware', 'r')
        if not fh then return 'Unknown' end
        local last_line
        for line in fh:lines() do
            if line:find('FPGA') then last_line = line end
        end
        fh:close()
        if not last_line then return 'Unknown' end
        local version = last_line:match('%[FPGA%-(.+)%]')
        fh =
            io.popen('spi-flash-tool verify /usr/share/fpga-config/digcomm.bin')
        if fh then
            for line in fh:lines() do
                if line:find('Verified') then
                    fh:close()
                    return version or 'Unknown', true
                end
                break
            end
            fh:close()
        end
        return version or 'Unknown'
    end
}

return get_fpga_version[config.device_type] or function() return "Unknown" end
