local ffi = require 'ffi'
local C = ffi.C

ffi.cdef[[
    typedef unsigned long       fsblkcnt_t;
    typedef unsigned long       fsfilcnt_t;

    struct statvfs {
        unsigned long   f_bsize;        // file system block size
        unsigned long   f_frsize;       // fragment size
        fsblkcnt_t      f_blocks;       // size of fs in f_frsize units
        fsblkcnt_t      f_bfree;        // number of free blocks
        fsblkcnt_t      f_bavail;       // number of free blocks for unpriv user
        fsfilcnt_t      f_files;        // number of inodes
        fsfilcnt_t      f_ffree;        // number of free inodes
        fsfilcnt_t      f_favail;       // number of free inodes for unpriv user
        unsigned long   f_fsid;         // file system ID
        int             __f_unused;     // Unused
        unsigned long   f_flag;         // mount flags
        unsigned long   f_namemax;      // Max filename length
        int             __f_spare[6];   // Padding
    };

    int statvfs(const char *path, struct statvfs *buf);
    char *strerror(int errnum);
]]

local m = {}

function m.get_available(path)
    local stat = ffi.new('struct statvfs[1]')
    local status = C.statvfs(path, stat)
    if status == 0 then
        return tonumber(stat[0].f_bavail) * tonumber(stat[0].f_frsize)
    end
    return nil, ffi.string(C.strerror(ffi.errno()))
end

function m.get_size(path)
    local stat = ffi.new('struct statvfs[1]')
    local status = C.statvfs(path, stat)
    if status == 0 then
        return tonumber(stat[0].f_blocks) * tonumber(stat[0].f_frsize)
    end
    return nil, ffi.string(C.strerror(ffi.errno()))
end

function m.get_storage_available()
    return m.get_available('/storage/')
end

function m.get_root_available()
    return m.get_available('/')
end

function m.get_storage_size()
    return m.get_size('/storage')
end

function m.get_root_size()
    return m.get_size('/')
end

return m
