local middleclass   = require 'middleclass';
local ffi           = require 'ffi'
local libhbmstream  = ffi.load('/usr/lib/libhbmstream.so')

ffi.cdef[[
    typedef struct hbmstream hbmstream_t;
    hbmstream_t *hbmstream_connect(struct ev_loop *loop, const char *ip);
    void hbmstream_add_subscription(hbmstream_t *stream,
        const char *signal_ref, int fd);
    void hbmstream_inhibit_watchdog(hbmstream_t *stream);
    void hbmstream_remove_subscription(hbmstream_t *stream,
        const char *signal_ref);
    void hbmstream_update_subscriptions(hbmstream_t *stream);
    void hbmstream_close(hbmstream_t *stream);
    int hbmstream_channels_available(hbmstream_t *stream);
]]

local hbmstream = class('hbmstream')

function hbmstream:initialize(loop, host)
    self.stream = libhbmstream.hbmstream_connect(loop, host)
end

function hbmstream:add_subscription(ref, fd)
    libhbmstream.hbmstream_add_subscription(self.stream, ref, fd)
end

function hbmstream:inhibit_watchdog()
    libhbmstream.hbmstream_inhibit_watchdog(self.stream)
end

function hbmstream:remove_subscription(ref)
    libhbmstream.hbmstream_remove_subscription(self.stream, ref)
end

function hbmstream:update_subscriptions()
    libhbmstream.hbmstream_update_subscriptions(self.stream)
end

function hbmstream:close()
    libhbmstream.hbmstream_close(self.stream)
    self.stream = nil
end

function hbmstream:channels_available()
    return libhbmstream.hbmstream_channels_available(self.stream)
end

return hbmstream
