local ffi = require'ptpv2-ffi/cdata'
local C = ffi.C

local m = {}

local function tx_ts_cb(ts, raw)
end

function m:send(time)
    local ci = self.clock_info
    local size = ci.sizeof_delay_resp_msg

    local th = ci.msg_hdr_save
    ffi.copy(th, ci.ibuf, ffi.sizeof('msg_hdr_t'))

    ffi.fill(ci.ibuf, size + 2)
    self:build_msg_header()
    local msg = ffi.cast('delay_resp_msg_t *', ci.ibuf)
    msg.hdr.transport_specific_message_type = C.PTP_MSG_TYPE_DELAY_RESP
    msg.hdr.message_length = C.htons(size)
    msg.hdr.domain_number = th.domain_number
    ffi.copy(msg.hdr.correction_field, th.correction_field,
        ffi.sizeof(msg.hdr.correction_field))
    msg.hdr.sequence_id = C.htons(th.sequence_id)
    msg.hdr.control_field = 0x3
    msg.hdr.log_message_interval = ci.log_min_delay_req_interval

    local rrts = time:to_timestamp()
    msg.receive_timestamp.seconds.msb = C.htons(rrts.seconds.msb)
    msg.receive_timestamp.seconds.lsb = C.htonl(rrts.seconds.lsb)
    msg.receive_timestamp.nanoseconds = C.htonl(rrts.nanoseconds)

    ffi.copy(msg.requesting_port_identity.clock_identity,
        th.source_port_identity.clock_identity,
        ffi.sizeof(msg.requesting_port_identity.clock_identity))
    msg.requesting_port_identity.port_number =
        C.htons(th.source_port_identity.port_number)

    if ci.transport_is_ipv6 ~= 0 then
        size = size + 2
    end

    local sock = self[C.GENERAL_PORT]
    local r, e = pcall(sock.sendto, sock, ci.ibuf, size, tx_ts_cb)
    if not r then
        self:log("error",
            string.format("send_delay_resp(): sendto() failed: %s", e))
        self:to_state(C.PTP_STATE_FAULTY)
        return
    end
    ci.sent_delay_req_sequence_id = ci.sent_delay_req_sequence_id + 1
end

function m:faulty(time, from_me)
    -- FIXME: Try to recover from faulty state?
end

function m:is_associated_with_delay_req(ci, msg)
    local expected_sid = ci.sent_delay_req_sequence_id - 1
    if expected_sid == msg.hdr.sequence_id and
        ci.port_identity.port_number ==
            C.ntohs(msg.requesting_port_identity.port_number) and
        C.memcmp(ci.port_identity.clock_identity,
            msg.requesting_port_identity.clock_identity,
            ffi.sizeof(ci.port_identity.clock_identity) == 0) then
        return true
    end
end

function m:uncalibrated(time, from_me)
    if self:is_from_current_master() then
        local ci = self.clock_info
        local msg = ffi.cast('delay_resp_msg_t *', ci.ibuf)
        if m.is_associated_with_delay_req(self, ci, msg) then
            msg.receive_timestamp.seconds.msb =
                C.ntohs(msg.receive_timestamp.seconds.msb)
            msg.receive_timestamp.seconds.lsb =
                C.ntohl(msg.receive_timestamp.seconds.lsb)
            msg.receive_timestamp.nanoseconds =
                C.ntohl(msg.receive_timestamp.nanoseconds)
            self.delay_req_receive_time =
                self.timing_iface:to_internal_time(msg.receive_timestamp)
            local cf =
                self.timing_iface:i64_to_internal_time(msg.hdr.correction_field)
            self:update_delay(cf)
            ci.log_min_delay_req_interval = msg.hdr.log_message_interval
        end
    end
end

m.handle_state = {}
m.handle_state[C.PTP_STATE_FAULTY] = m.faulty
m.handle_state[C.PTP_STATE_UNCALIBRATED] = m.uncalibrated
m.handle_state[C.PTP_STATE_SLAVE] = m.handle_state[C.PTP_STATE_UNCALIBRATED]

function m:handle(time, from_me)
    local state = self.clock_info.state
    local f = m.handle_state[state]
    if f then
        f(self, time, from_me)
    end
end

return m
