local armadb_client = require 'armadb_client'
local network_db = require 'network_db'
local cjson = require 'cjson.safe'

local m = {}

local client = armadb_client:new({})

local function get_setup(ip, id)
    if client then
        client.host = ip
        client:set({
            query = {
                authorization_key = 'locally_authorized',
                author = 'admin',
            }
        })
        local ok, res = pcall(client.request, client, 'GET',
            '/api/tracked_db/setups/' .. id .. '/_all_docs')
        if ok and not res.error then
            return cjson.encode(res)
        end
    end
end

-- Get the ZCIP addresses of all of the XR devices EXCEPT me!  We have to use
-- the ZCIP addresses because the curl stuff used doesn't seem to be able to
-- connect via IPv6!
local function get_remote_xr_device_ips()
    local ips = {}
    local ndb = network_db:new()
    local owner = ndb:get_inside_network_hostname()
    if not owner then
        ndb:close()
        return ips
    end
    local rrs = ndb:query([[
        SELECT rdata FROM unified WHERE type = 'TXT' AND local_source == '' AND
        owner != ? AND rdata LIKE 'platform_info:%' AND
        (owner LIKE 'cx23r%' OR owner LIKE 'edaq%');
    ]], owner)
    ndb:close()
    for _, rr in ipairs(rrs) do
        local t = cjson.decode(rr.rdata:match('^platform_info: (.*)$'))
        if t then
            ips[#ips + 1] = t.zcip_address
        end
    end
    return ips
end

-- Find and retrieve a setup from the remote XR devicess
function m.find_and_get_setup(id)
    local remote_ips = get_remote_xr_device_ips()
    for _, ip in ipairs(remote_ips) do
        local v = get_setup(ip, id)
        if v then return v end
    end
end

local function drill_down(t, key_list)
    local k = table.remove(key_list, 1)
    if #key_list == 0 then return t[k] end
    return drill_down(t[k], key_list)
end

local function get_preference(ip, key_list)
    if client then
        client.host = ip
        client:set({
            query = {
                authorization_key = 'locally_authorized',
                author = 'admin',
            }
        })
        local res
        local ok, res = pcall(client.request, client, 'GET',
            '/api/tracked_db/users/system/_all_docs')
        if ok and res and res.ok == nil then
            return drill_down(res[1], key_list)
        end
    end
end

local function duplicate_table(t)
    local newt = {}
    for k, v in pairs(t) do
        if type(v) == 'table' then
            newt[k] = duplicate_table(v)
        else
            newt[k] = v
        end
    end
    return newt
end

function m.is_preference_consistent(key_list)
    local kl = duplicate_table(key_list)
    local lp = get_preference('127.0.0.1', kl)
    if lp == nil then return end
    local remote_ips = get_remote_xr_device_ips()
    for _, ip in ipairs(remote_ips) do
        kl = duplicate_table(key_list)
        local v = get_preference(ip, kl)
        if v ~= nil and v ~= lp then return end
    end
    return true
end

--local v = m.find_and_get_setup('d5741b3a-1f2a-40ce-bb78-c42ea7c90822')
--print(v)

--local v = m.is_preference_consistent({'system', 'teds_enabled'})
--print(v)

return m
