"use strict";
/*global
    uuid
    Slick
    server_session
    EventSource
    alert
*/

$(function() {

    var app = window.app;
    var ch_data_view;
    var ch_grid;
    var ch_spreadsheet;
    var db_id = null;
    var grid_live_readings = {};
    var hw_spreadsheet;
    var live_cache = {};
    var panel_displays = {};
    var setup_menu;
    var comet;
    var comet_opts;
    var target_queue = $();
    var channel_search_queue = [];
    var session_key_arg = 'session_key=' +
        encodeURI(server_session.session_key);

    app.ui_state = {};

    var edit_menu_items = [
        {action: "cut", label_tr: "Cut", disabled: true},
        {action: "copy", label_tr: "Copy", disabled: true},
        {action: "paste", label_tr: "Paste", disabled: true},
        {action: "delete", label_tr: "Delete", disabled: false}
    ];

    var armadaProfiles = {
        ArmadaCheckbox: {
            editor: Slick.Editors.ArmadaCheckbox,
            formatter: Slick.Formatters.ArmadaCheckbox
        },
        ArmadaLive: {
            editor: null, // KLUDGE?
            formatter: Slick.Formatters.ArmadaLive
        },
        default: {
            editor: Slick.Editors.Text, // KLUDGE until autocomplete
            formatter: null
        }
    };
    var armadaSlickFactory = {
        getEditor: function(column) {
            return armadaProfiles[column.profile || 'default'].editor;
        },
        getFormatter: function(column) {
            return armadaProfiles[column.profile || 'default'].formatter;
        }
    };
    var ch_options = {
        editable: true,
        asyncEditorLoading: true,
        autoEdit: true,
        enableAddRow: true,
        enableCellNavigation: true,
        enableColumnReorder: true,
        syncColumnCellResize: true,
        editorFactory: armadaSlickFactory,
        formatterFactory: armadaSlickFactory,
        //autoHeight: true
        autoHeight: false
    };

    function dialog(message) {
        var title = "Notice";
        var content = $(
            '<div title="' + title + '">' +
            '<p>' + message + '</p>' +
            '</div>'
        );
        content.dialog({
            modal: true,
            buttons: {
                Ok: function () {
                    $(this).dialog('close');
                }
            }
        });
    }

    function make_id(ch_name) {
        if (typeof(ch_name) !== "string")
            return "";
        var id = ch_name.split(/[\.\@]/).reverse().join('.');
        id = id.replace(/[^_a-zA-Z0-9]/g, '_');
        id = id.replace(/([0-9]+)/g, function(s) {
            var zeros = "00000000000000000000";
            return zeros.substring(0, 10 - s.length) + s;
        });
        return id;
    }

    function get_hw_from_id(id) {
        if (id === 'source/expression') {
            return {
                name: 'computed_source',
                description: 'Computed source',
                category: 'source/expression',
                sample_rate: 10
            };
        } else if (id === 'filter/expression') {
            return {
                name: 'computed_channel',
                description: 'Computed channel',
                category: 'filter/expression',
                sample_rate: 0
            };
        } else if (id === 'sink/time_history') {
            return {
                name: 'time_history',
                description: 'Time history',
                category: 'sink/time_history',
                trigger_mode: 'always_on',
                sample_rate: 0
            };
        } else {
            var hw_view = hw_spreadsheet.armada_grid('innards', 'view');
            var in_ch = hw_view.getItemById(id);
            return in_ch;
        }
    }

    function add_panel_display(connector, id, f) {
        if (grid_live_readings[connector] == null)
            grid_live_readings[connector] = {};

        if (panel_displays[connector] == null)
            panel_displays[connector] = {};
        panel_displays[connector][id] = f;
    }

    function add_spreadsheet_item(ss, ch, idx) {
        var view = ss.armada_grid('innards', 'view');

        if (ch.id == null)
            ch.id = uuid();
        if (ch.reading == null)
            ch.reading = '&#8211;'; // en-dash
        if (ch.collect == null)
            ch.collect = false; // KLUDGE WRONG

        var sort_id = ch.sort_id = make_id(ch.connector); // KLUDGE

        if (idx == null) {
            idx = 0;
            var len = view.getLength();
            var items = view.getItems();
            for (var i = 0; i < len; i++) {
                if (items[i].sort_id < sort_id)
                    ++idx;
            }
        }
        view.insertItem(idx, ch);

        if (typeof(ch.connector) === "string" && ch.connector.length > 0) {
            (function(grid, connector, id){
                // KLUDGE need to be able to remove display
                add_panel_display(connector, id, function(v){
                    ch.reading = v;
                    var alr = grid_live_readings[connector][id];
                    if (alr !== undefined)
                        grid.updateCell(alr[0], alr[1]);
                });
            }(ss.armada_grid('innards', 'grid'), ch.connector, ch.id));
        }
    }

    function remove_spreadsheet_item(ss, ch) {
        // KLUDGE this is VERY hardware list specific
        var view = ss.armada_grid('innards', 'view');
        var len = view.getLength();
        var items = view.getItems();
        for (var i = 0; i < len; i++) {
            // KLUDGE should check category = 'hardware' or something
            if (items[i] != null && items[i].connector === ch.connector) {
                var item = items[i];
                var alr = grid_live_readings[item.connector];
                if (alr != null && alr[item.id] != null)
                    delete alr[item.id];
                view.deleteItem(item.id);
            }
        }
    }

    function remove_display_row(ch_name) {
        var ch = live_cache[ch_name];
        if (ch) {
            ch.jq.closest(".chdisplay").remove();
            live_cache[ch_name] = null;
        }
    }

    function add_files_row(v, label, value) {
        var ch = live_cache[v.id];
        if (ch != null)
            return;
        var id = make_id(v.id);
        var append_idx = 0;
        $("#files [id]").each(function(i) {
            if (this.id < id)   // KLUDGE sort by date
                ++append_idx;
        });
        $("#files tr").eq(append_idx).after(
            '<tr class="chdisplay">' +
            '<td><a href="/storage/' + v.name + '">' + label + '</a></td>' +
            '<td id="' + id + '">' + value + '</td>' +
            '</tr>'
        );
        var id_jq = $("#" + id);
        live_cache[v.id] = {jq: id_jq, id: id};
        id_jq.find("a").click(function(e) {
            $.get($(this).attr("href"), null, function(data) {
                dialog("Deleted " + v.name); // KLUDGE do error handling
            });
            return false;
        });
    }

    function add_table_row(table_id, v, cols) {
        var ch = live_cache[v.id];
        if (ch != null)
            return;
        var id = make_id(v.id);
        var append_idx = 0;
        $(table_id + " [id]").each(function(i) {
            if (this.id < id)   // KLUDGE sort by date
                ++append_idx;
        });
        var tr = $('<tr class="chdisplay">' + cols + '</tr>');
        $(table_id + " tr").eq(append_idx).after(tr);
        var id_jq = $("#" + id);
        live_cache[v.id] = {jq: id_jq, id: id};
        return tr;
    }

    function add_draggable(ch) {
        var draggable_ch = $('<p>' + ch.description + '</p>').draggable({
            opacity: 0.7,
            helper: (function(ch) {
                return function(ev) {
                    return $("<div></div>")
                        .text(ch.description)
                        .append("<span></span>")
                        .addClass("ui-widget armada-draggable")
                        .appendTo("body");
                };
            }(ch)),
            /*
            // For drag/drop target position debugging
            drag: function(ev, ui) {
                var off = ui.offset;
                var ss_off = ch_spreadsheet.offset();
                var pos = {
                    top: off.top - ss_off.top,
                    left: off.left - ss_off.left
                };
                var grid = ch_spreadsheet.armada_grid('innards', 'grid');
                var cell = grid.getCellFromPoint(pos.left, pos.top);
                var text =
                    ":<br>\n"
                    + "off=" + $.toJSON(off) + "<br>\n"
                    + "ss_off=" + $.toJSON(ss_off) + "<br>\n"
                    + "pos=" + $.toJSON(pos) + "<br>\n"
                    + "cell=" + $.toJSON(cell);
                $("span", ui.helper).html(text);
            },
            */
            revert: true,
            revertDuration: 0
            // containment: "#displays-tab" // KLUDGE WRONG
        }).data('armada-id', ch.id).data('armada-sort-id', ch.sort_id);
        if (ch.connector != null)
            draggable_ch.data('armada-connector', ch.connector);
        $('#available-channels p').each(function(){
            if ($(this).data('armada-sort-id') > ch.sort_id) {
                draggable_ch.insertBefore($(this));
                draggable_ch = null;
                return false;
            }
        });
        if (draggable_ch)
            $('#available-channels').append(draggable_ch);
    }

    var updaters = {
        '#ajax': function(v) {
            $.ajax(v);
        },
        '#live_data' : function(v) {
            for (var ch_name in v) {
                var reading = v[ch_name];
                var pd = panel_displays[ch_name];
                if (pd != null) {
                    for (var id in pd)
                        pd[id](reading, id);
                }
            }
        },
        '#add_connector' : function(e) {
            //add_display_item(e, "CONNECTOR: " + e, '&nbsp;');
        },
        '#remove_connector' : function(e) {
            //remove_spreadsheet_item(e);
        },
        '#add_channel' : function(channels) {
            for (var i = 0; i < channels.length; i++) {
                var ch = channels[i];
                if (ch.category == null)
                    ch.category = 'source';
                add_spreadsheet_item(hw_spreadsheet, ch);
                add_draggable(ch);
            }
        },
        '#remove_channel' : function(channels) {
            /*jshint loopfunc: true */
            for (var i = 0; i < channels.length; i++) {
                var ch = channels[i];
                remove_spreadsheet_item(hw_spreadsheet, ch);
                $('#available-channels p').each(function(){
                    if ($(this).data('armada-connector') === ch.connector) {
                        $(this).remove();
                        return false;
                    }
                });
            }
        },
        '#update_file' : function(v) {
            var item = live_cache[v.id];
            var status = v.size + ' bytes, ' + v.modified +
                ' <a href="/comet/delete_file/' + v.name + '">Delete</a>';
            if (item == null) {
                add_files_row(v, v.name, status);
            } else {
                item.jq.html(status);
            }
        },
        '#remove_file' : function(v) {
            remove_display_row(v);
        },
        '#update_storage' : function(v) {
            var item = live_cache[v.id];
            var status = v.id + ': ' + v.available + ' / ' + v.capacity +
                ' bytes free';
            if (item == null) {
                var id = make_id(v.id);
                add_table_row("#storage", v,
                    '<td id="' + id + '">' + status + '</td>');
            } else {
                item.jq.html(status);
            }
        },
        '#remove_storage' : function(v) {
            remove_display_row(v);
        },
        '#tracked_db_server_request': function(args, origin) {
            var url = '/comet/tracked_db/' + args.dir + '/' + args.doc +
                (args.item ? '/' + args.item : '') +
                (args.url_extras ? args.url_extras : '');
            $.ajax({
                type: args.type,
                url: url,
                cache: false,
                contentType: 'application/json',
                data: args.data ? $.toJSON(args.data) : null,
                success: args.success || function(data, status, xhr) {
                },
                error: args.error || function(data, status, xhr) {
                }
            });
        },
        '#tracked_db_create' : function(v) {
        },
        '#tracked_db_delete' : function(e) {
        },
        '#tracked_db_delta': function(v, origin) {
            // Don't propagate updates that came from this session.
            // Infinite loops are bad.
            // KLUDGE We may want to retire "outstanding" requests here,
            // which would be kept around until then so that edits that
            // don't make it to the (down, for example) server can get
            // re-sent once the server is reachable again.
            if (v.source_session_key === server_session.session_key)
                return; // KLUDGE retire outstanding edit request here?
            if (v.db !== db_id) {
                // KLUDGE Handle stuff like 'users', other tests 'summary', etc.
                return;
            }
            if (v.schema === 'setups') { // KLUDGE??? -- schema => "directory"
                $.command('#edit_delta',
                    $.extend({}, v, {origin: origin}), origin);
            }
        },
        '#create_setup' : function(v) {
            var setup_id = v.setup_id || uuid();
            ch_data_view.createBackendSetup(setup_id); // KLUDGE error handling
        },
        '#update_ui_state': function(args) {
            $.extend(app.ui_state, args);
            $.command('#tracked_db_server_request', {
                type: 'POST',
                dir: 'setup_user_states',
                doc: server_session.state.user.id, // KLUDGE WRONG user.uid
                item: db_id,
                data: app.ui_state
            });
        },
        '#open_setup' : function(v) {
            var new_id = v.setup_id;
            if (new_id === db_id)
                return; // KLUDGE error / message / ???
            var is_new = (new_id === '');
            if (is_new) {
                new_id = uuid();
                $.command('#create_setup', {setup_id: new_id});
            }

            db_id = new_id;
            ch_data_view.setBackend({
                url: '/comet/tracked_db/setups/',
                db_id: new_id
            });

            if (is_new) {
                $.command('#edit_delta', {new: {
                    summary: {
                        id: "summary",
                        category: "summary",
                        setup_id: new_id,
                        test_name: "new_setup",
                        test_title: "New Setup",
                        test_author: server_session.state.user.name
                    }
                }});
            }

            $.command('#update_ui_state', {setup_id: db_id});
            $.command('#refresh_wired_ui_elements');
            ch_spreadsheet.armada_grid('innards', 'grid').invalidate();
        },
        '#edit_delta': function(act, origin) {
            if (origin === 'remote')
                act = $.extend({}, act, {origin: origin});
            ch_data_view.deltaItems(act);
            // Edits should propagate through DataView onDelta event
        },
        '#update_tests' : function(v) {
            var item = live_cache[v.id];
            var status =
                '<input type="button" name="stop_test" value="Stop Test">' +
                    ' Started: ' + v['start-time'];
            if (item == null) {
                var id = make_id(v.id);
                var tr = add_table_row("#tests", v,
                    '<td>' + v.name + '</td>' +
                    '<td id="' + id + '">' + status + '</td>'
                );
                tr.find("input").data(v);
            } else {
                item.jq.html(status);
            }
        },
        '#remove_tests' : function(v) {
            remove_display_row(v);
        },
        '#start_test': function(test_name) {
            var setup = [];

            //dialog('Start Test button clicked');

            var row_count = ch_data_view.getLength();
            for (var i = 0; i < row_count; i++) {
                var item = ch_data_view.getItemByIdx(i);
                if (item)
                    setup.push(item);
            }

            if ($.isEmptyObject(setup)) {
                dialog("No channels are selected for the test");
                return false;
            }

            var json = $.toJSON(setup);

            //dialog("serialized: " + $.toJSON(form.serializeArray()))
            //dialog("start_test: " + json)

            $.ajax({
                type: 'POST',
                url: '/comet/start_test/' + test_name,
                cache: false,
                contentType: 'application/json',
                data: json,
                success: function(data, status, xhr) {
                    dialog("Test started collecting " + setup.length +
                        " channels");
                },
                error: function(data, status, xhr) {
                    dialog("Test NOT started");
                }
                // dataType: dataType
            });
            return false;
        },
        '#stop_test': function(v) {
            //dialog("stop_test: " + $.toJSON(v));
            $.ajax({
                type: 'GET',
                url: '/comet/stop_test/' + v.srv_port, // KLUDGE?
                cache: false,
                success: function(data, status, xhr) {
                    dialog("Stopping test");
                },
                error: function(data, status, xhr) {
                    dialog("Test NOT stopped");
                }
            });
            return false;
        },
        '#refresh_log': function() {
            $.ajax({
                type: 'GET',
                url: '/comet/get_log',
                cache: false,
                success: function(data, status, xhr) {
                    $('#log_body').html(data);
                }
            });
        },
        '#change_tab': function(tab) {
            app.tabs.tabs('option', 'active', '#' + tab);
        },
        '#rewire_ui_elements': function(args) {
            // KLUDGE need incremental version too
            app.wired_ui_elements = {};
            var wired = app.wired_ui_elements;
            $('*[class]').each(function() {
                var self = $(this);
                var classes = self.attr('class').split(' ');
                for (var i = 0, len = classes.length; i < len; ++i) {
                    var matches = classes[i].match(/^V-([^_]*)_(.*)$/);
                    if (matches) {
                        var id = matches[1], field = matches[2];
                        wired[id] = wired[id] || {};
                        wired[id][field] = wired[id][field] || [];
                        wired[id][field].push($(this)); // KLUDGE $(this)/self ?
                    }
                }
            });
            // KLUDGE (?) for title - want more general way to do this
            wired.summary = wired.summary || {};
            wired.summary.test_title = wired.summary.test_title || [];
            wired.summary.test_title.push(function(val) {
                $('title').text(val + ' - eDAQ Armada'); // KLUDGE TR()
            });
        },
        '#refresh_wired_ui_elements': function() {
            // Updates all wired elements
            var wired = app.wired_ui_elements;
            _.each(wired, function(fields, id) {
                var item = ch_data_view.getItemById(id) || {};
                _.each(fields, function(targets, field) {
                    var val = item[field] || '';
                    _.each(targets, function(w) {
                        if (typeof w === 'function') {
                            w(val);
                        } else if (w.is(':input')) {
                            var w_val = w.val();
                            if (w_val !== val) {
                                // KLUDGE flag old_val !== w_val conflict???
                                w.val(val);
                            }
                        } else {
                            w.text(val); // KLUDGE html???
                        }
                    });
                });
            });
        },
        '#update_wired_ui_elements': function(delta) {
            // Only updates wired elements found in delta.new
            // Does NOT update :input elements whose current value
            // does not match delta.old value, to avoid update/typing
            // races.
            var wired = app.wired_ui_elements;
            _.each(delta.new, function(d_fields, id) {
                var item_old = delta.old[id] || {};
                if (!wired[id])
                    return;
                _.each(d_fields, function(val, field) {
                    if (!wired[id][field])
                        return;
                    _.each(wired[id][field], function(w) {
                        if (typeof w === 'function') {
                            w(val);
                        } else if (w.is(':input')) {
                            var w_val = w.val();
                            var old_val = item_old[field];
                            if (w_val !== val && old_val === w_val) {
                                // KLUDGE flag old_val !== w_val conflict???
                                w.val(val);
                            }
                        } else {
                            w.text(val); // KLUDGE html???
                        }
                    });
                });
            });
        }
    };

    $.command('%register_commands', updaters);

    function menu_select(ev, ui) {
        if (ui.item.hasClass("ui-state-disabled"))
            return true;
        var action = ui.item.data("armada-action");
        $.command(action, {event: ev, ui: ui});
        return true;
    }

    function create_menu(spec) {
        var menu = $('<ul></ul>');
        for (var i = 0; i < spec.items.length; i++) {
            var item = spec.items[i];
            var ui_item = $('<li><a href="#">' + item.label_tr + '</a></li>');
            ui_item.data("armada-action", item.action);
            if (item.disabled)
                ui_item.addClass("ui-state-disabled");
            menu.append(ui_item);
        }
        menu.menu({
            select: spec.select || menu_select
        });
        $("body").append(menu); // KLUDGE!?
        return menu;
    }

    function make_readout(display_panel, type, ch, off) {
        var target = $('<div class="display-item"></div>');
        target.css("position", "absolute") // KLUDGE for chrome
            .offset(off)
            .draggable({stack: ".display-item"})
            .resizable({
                handles: 'all',
                //aspectRatio: 1,
                helper: "display-resizer"
            });

        display_panel.append(target);

        target[type]({
            skin: 'white',
            legend: ch.name,
            unit: ch.units,
            reading: ch.reading,
            range_min: Number(ch.range_min),
            range_max: Number(ch.range_max),
            height: 220,
            width: 200,
            left: off.left,
            top: off.top
        });
        target.bind('resizestop', function (ev, ui) {
            target[type]('set_configuration', ui.size);
        });
        add_panel_display(ch.connector, ch.id, function (v) {
            target[type]('set_reading', v);
        });

        target.dblclick(function (ev) {
            var readouts_menu = create_menu({items: [
                {action: "reading", label_tr: "Numeric"},
                {action: "dial", label_tr: "Dial"},
                {action: "bar", label_tr: "Bar Gauge"},
                {action: "strip_multi", label_tr: "Strip Chart"},
                {action: "", label_tr: ""},
                {action: "delete", label_tr: "Delete"}
            ], select: function (ev, ui) {
                //dialog("readouts_menu: " + ui.item.text() + " " + ui.item.data("armada-action") + " " + ui.item.hasClass("ui-state-disabled"));
                var new_type = ui.item.data("armada-action");
                if (new_type === "delete") {
                    target[type]('destroy');
                    target.remove();
                } else if (new_type !== "") {
                    target[type]('destroy');
                    target.unbind('dblclick');
                    target.unbind('resizestop');
                    off = target.offset();
                    target.remove();
                    make_readout(display_panel, "somat_" + new_type, ch, off);
                }
            }});
            readouts_menu.popup({
                trigger: $(this),
                role: "menu",
                position: {
                    my: "center",
                    at: "center"
                }
            }).popup('open', ev);
            readouts_menu.popup('option', {
                close: function (ev) {
                    setTimeout(function () {
                        readouts_menu.popup("destroy");
                        readouts_menu.menu("destroy");
                        readouts_menu.remove();
                    }, 1);
                }
            });
            return false;
        });
    }

    function drop_channel(event, ui) {
        /*jshint validthis: true */
        var is_display = ui.draggable.hasClass("display-item");
        if (is_display) {
            // Update database item here
        } else {
            // Add new display element here
            var in_ch = get_hw_from_id(ui.draggable.data('armada-id'));
            var ch = $.extend({}, in_ch, {id: uuid()});
            var off = ui.offset;

            make_readout($(this), "somat_dial", ch, off);

            // Create database item here
        }
    }

    function log(m) {
        $('#debug-tab').append('<p>' + m + '</p>');
    }

    $("#displays-tab").droppable({
        drop: drop_channel
    });
    $('#common_gauge_definitions').svg({
        loadURL: 'js/armada_readouts/svg/definitions.svg'
    });
    setup_menu = $('#test_name_auto');
    function or_nil(s) {
        return (s == null) ? "" : s;
    }
    setup_menu.autocomplete({
        minLength: 0,
        select: function(ev, ui) {
            $.command('#open_setup', {setup_id: ui.item.value});
        },
        source: function(request, response) {
            ch_data_view.backendRequest({
                type: 'GET',
                url: '/comet/tracked_db/setups/_all_dbs?match_id=summary',
                success: function(data, status, xhr) {
                    var menu = [];
                    $.each(data, function(k, v) {
                        var test_name = or_nil(v.test_name) + ' "' +
                            or_nil(v.test_title) + '" (' +
                            or_nil(v.test_author) + ')';
                        menu.push({value: k, label: test_name});
                    });
                    menu.sort(function (a, b) {
                        return a.label.localeCompare(b.label);
                    });
                    menu.unshift({value: '', label: 'New Setup...'});
                    response(menu);
                }
            });
        }
    });
    // KLUDGE invisible-autocomplete style doesn't set these firmly enough...
    setup_menu.css({width: 1, height: 1});
    $('#header_setup_selector').click(function() {
        var auto = $('#test_name_auto');
        if (auto.autocomplete('widget').is(':visible')) {
            auto.autocomplete('close');
            return;
        }
        auto.autocomplete('search', ''); // KLUDGE WRONG
        auto.focus();
    });

    var tabs = app.tabs = $("#tabs");
    var tab_children = {};
    $("#tabs-repo").children().each(function(idx, el){
        var id = $(this).attr('id');
        tab_children[id] = $(this).detach();
    });
    var tab_html = '';
    var first_tab = null;
    var tab_spec = app.defaults.tabs;
    for (var i = 0; i < tab_spec.length; i++) {
        var ts = tab_spec[i];
        tab_html += '<li id="tab_' + i + '"><a href="#' + ts.target + '"' +
            ' data-tab_name="' + ts.target + '">' +
            ts.label + '</a></li>';
        tabs.append(tab_children[ts.target]);
        if (first_tab == null)
            first_tab = tab_children[ts.target];
        delete tab_children[ts.target];
    }
    tab_children = null;
    tabs.prepend('<ul>' + tab_html + '</ul>');
    tabs.tabs({
        //beforeload: function() { }
        activate: function(event, event_data) {
            $.command('#update_ui_state', {
                current_tab: event_data.newTab.data('tab_name')
            });
        }
    });
    tabs.find(".ui-tabs-nav").sortable({
        axis: "x",
        distance: 5,
        update: function(e, ui){
            var msg = "";
            //$(">div", tabs).each(function(i, e){ msg += $(e).attr('id') + ", "; });
            //log("Tabs reordered: " + $(this).sortable('serialize'));
            //msg = $(this).sortable('toArray', {attribute:"id"}).join(", ");
            msg = $(this).sortable('toArray').join(", ");
            //log("Tabs reorderedXXX: " + msg);
        }
    });

    function resizer() {
        var panel = $('#panel');
        var h = panel.innerHeight() - first_tab.offset().top - 28;
        var w = panel.innerWidth() - 40;
        $('.tab').each(function(){
            var t = $(this);
            if (t.height() < h)
                t.height(h);
        });
        $('.display-tab').css({width: w, height: h});

        var new_size = {height: h - 100};
        ch_spreadsheet
            .css(new_size)
            .armada_grid('innards', 'grid').resizeCanvas();

        hw_spreadsheet
            .css(new_size)
            .armada_grid('innards', 'grid').resizeCanvas();
    }

    $(window).resize(resizer);

    $.ajaxPrefilter(function(opts, orig_opts, xhr) {
        // KLUDGE there might be a better way to do this...?
        var sep = (opts.url.match(/\?/)) ? '&' : '?';
        opts.url = opts.url + sep + session_key_arg;
    });

    // KLUDGE working prototype of instant change propagation from at
    // least some classes of input fields on typing and click events
    // KLUDGE maybe jquery.instant_edit.js or some such eventually?

    function process_target_queue() {
        var delta = {};
        var date = new Date();
        var stamp = date.getUTCSeconds() + date.getUTCMilliseconds() * 0.001;

        target_queue.each(function() {
            var target = $(this);
            var item_id = target.closest(".item-id-ref").data('item_id');
            var name = target.attr('name');
            var val = target.val();
            target.data('timestamp', stamp);
            if (item_id) {
                if (delta[item_id] == null)
                    delta[item_id] = {id: item_id};
                delta[item_id][name] = val;
            }
        });
        if (!$.isEmptyObject(delta)) {
            $.command('#edit_delta', {
                timestamp: stamp,
                new: delta
            });
        }
        target_queue = $();
    }

    function queue_target(ev) {
        var target = $(ev.target);
        // KLUDGE hasClass('wired') or some such???
        // KLUDGE also might want to relax the :input spec if doing crazy
        // things like trying to manually edit a <div> or something
        if (target.is(':input')) {
            var date = new Date();
            var stamp = date.getUTCSeconds() +
                date.getUTCMilliseconds() * 0.001;
            target.data('timestamp', stamp);
            var class_attr = target.attr('class');
            if (class_attr && !!class_attr.match(/\bV-/)) {
                // Defer all of our processing until after all the event's
                // effects have percolated through the system
                if (target_queue.length === 0)
                    setTimeout(process_target_queue, 0);
                if (!target_queue.is(target))
                    target_queue.push(target);
            }
        }
        return true;
    }

    // KLUDGE use $.on() for jQuery >= 1.7
    $('body').bind('keydown keyup mousedown mouseup focusin focusout',
        queue_target);

    $("input[name=start_test]").live('click', function(e) {
        var test_name = $("input[name=test_name]").val();
        $.command('#start_test', test_name);
        return false;
    });
    $("input[name=stop_test]").live('click', function(e) {
        $.command('#stop_test', $(this).data());
        return false;
    });
    $("input[name=refresh_log]").live('click', function(e) {
        $.command('#refresh_log');
        return false;
    });

    // KLUDGE clone channel_columns here, or better, use it unmodified
    $.each(app.defaults.channel_columns, function(i, col) {
        if (col._live_key)
            col._live_readings = grid_live_readings; // KLUDGE!
    });

    hw_spreadsheet = $("#hardware-list").armada_grid({
        columns: app.defaults.channel_columns,
        slickgrid_options: ch_options
    });
    hw_spreadsheet.armada_grid('option', 'column_set',
        {label: "Hardware", columns:  [
            {id: "connector"},
            {id: "description"},
            {id: "reading"},
            {id: "units"},
            {id: "output_state"},
            {id: "unit_identify"}
        ]}
    );
    ch_spreadsheet = $("#ch_slick").armada_grid({
        columns: app.defaults.channel_columns,
        slickgrid_options: $.extend({}, ch_options, {
            data_view: new Slick.Data.ArmadaSetupModel({
            })
        }),
        validate: true
    });
    ch_data_view = ch_spreadsheet.armada_grid('innards', 'view');
    ch_data_view.onBackendChanged.subscribe(function(e, args) {
        $.command('#refresh_wired_ui_elements');
    });
    ch_data_view.onRowsChanged.subscribe(function(e, args) {
    });
    ch_data_view.onRowCountChanged.subscribe(function(e, args) {
        $.command('#refresh_wired_ui_elements');
    });
    ch_data_view.onDelta.subscribe(function(e, delta) {
        $.command('#update_wired_ui_elements', delta, "immediate");
    });
    ch_data_view.setBackend({
        url: '/comet/tracked_db/setups/',
        db_id: db_id
    });
    ch_grid = ch_spreadsheet.armada_grid('innards', 'grid');

    ch_grid.onAddNewRow.subscribe(function(e, args) {
        add_spreadsheet_item(ch_spreadsheet, args.item,
            ch_data_view.getLength());
    });

    ch_spreadsheet.droppable({
        drop: function(event, ui) {
            var ss = $(this);
            var grid = ss.armada_grid('innards', 'grid');
            var in_ch = get_hw_from_id(ui.draggable.data('armada-id'));
            if (in_ch == null)
                return true;

            var off = ui.offset;
            var ss_off = ss.offset();
            var pos = {
                top: off.top - ss_off.top,
                left: off.left - ss_off.left
            };
            var cell = grid.getCellFromPoint(pos.left, pos.top);
            if (cell.row < 1)
                cell.row = 1;
            else if (cell.row > ch_data_view.getLength() + 1)
                cell.row = ch_data_view.getLength() + 1;

            var new_ch = $.extend({}, in_ch, {id: uuid(), collect: true});
            add_spreadsheet_item(ss, new_ch, cell.row - 1);
        }
    });

    $.each(app.defaults.channel_column_sets, function(i, col) {
        var control = $('<button>' + col.label + '</button>');

        control.click((function(c){
            return function(){
                ch_spreadsheet.armada_grid('option', 'column_set', c);
                $(this).parent().find("button[checked]").attr('checked',
                    false);
                $(this).attr('checked', true);
            };
        }(col)));

        // KLUDGE hardcoded wrong! - (from user state, not test?)
        if (i === 0) {
            control.click();
        }
        $("#spreadsheet_column_selectors").append(control);
    });

    $("#spreadsheet_filter").keyup(function(ev){
        //log("Filter: " + $(this).val());
        ch_spreadsheet.armada_grid('option', 'filter_string',
            $(this).val());
        // KLUDGE return value?
    });

    $(".refresh-button").click(function(){
        ch_data_view.refresh();
        ch_grid.render();
        //dialog("refresh");
    });

    $("#available-channels")
        .addClass("ui-menu ui-widget ui-widget-content ui-corner-all")
        .hide();

    add_draggable({
        id: "source/expression",
        description: "Computed source",
        sort_id: "0"
    });
    add_draggable({
        id: "filter/expression",
        description: "Computed channel",
        sort_id: "1"
    });
    add_draggable({
        id: "sink/time_history",
        description: "Time history",
        sort_id: "2"
    });

    function channel_search_event(ev) {
        var target = $(ev.target);
        var menu = $('#available-channels');

        switch (ev.type) {
        case 'focusin':
            menu.show().position({
                my: "right top",
                at: "right bottom",
                of: target
            });
            break;
        case 'focusout':
            menu.hide();
            break;
        case 'mousedown':
        case 'mouseup':
            break;
        default:
            break;
        }
        // Defer all of our processing until after all the event's
        // effects have percolated through the system
        if (channel_search_queue.length === 0) {
            channel_search_queue.push(target); // Only once
            setTimeout(function () {
                $.each(channel_search_queue, function(i, target) {
                    var val = target.val();
                    if (target.data('val') !== val) {
                        target.data('val', val);
                        // Now filter stuff
                        var re = new RegExp(val, 'i');
                        $('> p', menu).each(function(idx, el) {
                            var visible =
                                val === '' || $(this).text().match(re);
                            $(this)[visible ? 'show' : 'hide']();
                        });
                    }
                });
                channel_search_queue = [];
            }, 0);
        }
        return true;
    }

    $('.search-container').addClass("ui-widget");
    $('#channel-search')
        .addClass("ui-widget ui-widget-content ui-state-default")
        // KLUDGE use $.on() for jQuery >= 1.7
        .bind('keydown keyup mousedown mouseup focusin focusout',
            channel_search_event);

    create_menu({items: edit_menu_items}).popup({
        trigger: $("a[href=#edit-menu]"),
        role: "menu"
    });

    comet_opts = {
        url: "/comet/stream.comet?" + session_key_arg,
        dispatcher: $.command,
        global_comet_function_name: "comet_update",
        rate: 5
    };

    if (false) {
        /*jshint white: false */
        var fake_setup = {
            "update_rate": 5,
            "fake_preamble": [
                ["#add_connector", "gps-1.fake"],
                ["#add_channel", [{
                        "name": "seconds",
                        "description": "Faked GPS Decimal Seconds",
                        "label": "Seconds",
                        "range_min": "0",
                        "range_max": "60",
                        "sample_rate": "5",
                        "units": "s",
                        "connector": "seconds@gps-1.fake"
                    }, {
                        "name": "minutes",
                        "description": "Faked GPS Minutes",
                        "label": "Minutes",
                        "range_min": "0",
                        "range_max": "59",
                        "sample_rate": "5",
                        "units": "minutes",
                        "connector": "minutes@gps-1.fake"
                    }, {
                        "name": "hours",
                        "description": "Faked GPS Hours",
                        "label": "Hours",
                        "range_min": "0",
                        "range_max": "23",
                        "sample_rate": "5",
                        "units": "hours",
                        "connector": "hours@gps-1.fake"
                    }, {
                        "name": "days",
                        "description": "Faked GPS Days",
                        "label": "Days",
                        "range_min": "0",
                        "range_max": "1000",
                        "sample_rate": "5",
                        "units": "days",
                        "connector": "days@gps-1.fake"
                    }]
                ]
            ],
            "fake_live_data": "function(t, stamp) " +
                "local mod, floor = math.mod, math.floor " +
                "t['days@gps-1.fake'] = floor(stamp / (60 * 60 * 24)) " +
                "t['hours@gps-1.fake'] = mod(floor(stamp / (60 * 60)), 24) " +
                "t['minutes@gps-1.fake'] = mod(floor(stamp / 60), 60) " +
                "t['seconds@gps-1.fake'] = mod(stamp, 60) " +
            "end"
        };
        $.ajax({
            type: 'POST',
            url: "/comet/stream.eventsource?" + session_key_arg,
            cache: false,
            contentType: 'application/json',
            data: $.toJSON(fake_setup),
            error: function(data, status, xhr) {
                alert("FAILURE setting up fake data");
            }
        });
    }

    if (location.hash === "#simulator") {
        comet = $("body").armada_simulator(comet_opts);
        comet.armada_simulator('start');
        $('#comet-control').text("Simulator Running").click(function(ev){
            comet.armada_simulator('stop');
            $('#comet-control').text("Simulator Stopped");
        });
    } else if (location.hash !== "#comet" &&
            typeof EventSource !== 'undefined') {
        // KLUDGE - timeout to force event source to not hold page
        // in loading status - see https://groups.google.com/a/chromium.org/group/chromium-html5/browse_thread/thread/be1a6ee5b0af3a3d/0ac967234322ccca
        // If page is still loading iPad can't resize table.
        setTimeout(function () {
            var url = "/comet/stream.eventsource?" + session_key_arg;
            var eventsource = new EventSource(url);
            eventsource.onopen = function (event) {
                var stop_link = $('<span>(Stop)</span>')
                    .css('cursor', 'pointer')
                    .click(function () {
                        eventsource.close();
                        $('#comet-control')
                            .text(location.hostname + ' Stopped');
                        return false;
                    });
                $('#comet-control').text(location.hostname + ' ')
                    .append(stop_link);
            };
            eventsource.onerror = function (event) {
                $('#comet-control').text(location.hostname + ' Stopped');
            };
            eventsource.onmessage = function (event) {
                var e = $.parseJSON(event.data);
                for (var i = 0, len = e.length; i < len; i++) {
                    $.command(e[i][0], e[i][1], 'remote');
                }
            };
        }, 1000);
    } else {
        comet = $("body").armada_comet(comet_opts);
        comet.armada_comet('start');
        $('#comet-control').text(location.hostname);
    }

    resizer();

    $.command('#tracked_db_server_request', {
        type: 'GET',
        dir: 'setup_user_states',
        doc: server_session.state.user.id, // KLUDGE WRONG user.uid
        item: '_most_recent',
        success: function(data, status, xhr) {
            if (typeof(data) !== 'object' || typeof(data[0]) !== 'object')
                return; // KLUDGE error?
            var state = data[0];
            if (state.setup_id != null)
                $.command('#open_setup', {setup_id: state.setup_id});
            if (state.current_tab != null) {
                $.command('#change_tab', state.current_tab);
            }
        }
    });

    $.command('#rewire_ui_elements', {});
    $.command('#refresh_wired_ui_elements');
});
