/*jshint white:false */
(function( $, undefined ) {

$.widget( 'ui.somat_strip_multi',
{
    /**
     * @description This attribute defines the natural pixel width of the
     *              gauge. It is used for scale math.
     */
    _normalW     :  200,
    /**
     * @description This attribute defines the natural pixel height of the
     *              gauge. It is used for scale math.
     */
    _normalH     :   60,
    /**
     * @description This attribute represents the width in pixels of the
     *              gauges needle display. It is use for percentage math.
     */
    _meterFactor :   31,
    /**
     * @description This attribute respresents the natural pixel width of
     *              the gauges display. It is used for line factor math.
     */
    _lineBase    :  146,

    _baseURL     :  $('script:last').prop('src').replace(/[^\/]*$/, ''),

    //private
    'ready'             :      false,
    'scale_x'           :        1.0,
    'scale_y'           :        1.0,
    'range'             :        100,

    //layout
    'lowerX'            :        195,
    'lowerY'            :         40,
    'upperX'            :        195,
    'upperY'            :         15,
    'unitX'             :        190,
    'unitY'             :       27.5,
    'legendX'           :        100,
    'legendY'           :         56,
    'rangeFont'         :         40,
    'unitFont'          :         50,
    'legendFont'        :         40,
    //dom
    'svg'               :       null,
    'ticks1'            :       null,
    'ticks2'            :       null,


    /**
     * @description This object stores the instance data of this widget.
     * NOTES:
     *        channels    : 1 through 6, determined by gauge.css and
     *                      definitions.svg for coloration.
     *        mode        : 'linear' / 'min_max'
     *        pace        : time in milliseconds to scroll 1 reading
     *        show_needle : ignored when mode = 'min_max'
     *        filled      : ignored when mode = 'min_max'
     */
    options: {
        'skin'              :  'white',
        'show_needle'       :     true,
        'filled'            :    false,
        'keep_full'         :     true,
        'true_pixels'       :     true,
        'mode'              : 'linear',
        'channels'          :        1,
        'pace'              :       25,
        'top'               :        0,
        'left'              :        0,
        'height'            :       60,
        'width'             :      200,
        'legend'            :  'Label',
        'unit'              :    'psi',
        'range_min'         :        0,
        'range_max'         :      100,
        'channel_readings'  :     null,
        'ticks_major'       :     null,
        'ticks_minor'       :     null,
        'readout_range_min' :     null,
        'readout_range_max' :     null,
        'timer'             :     null,
        'resolution'        :      150,
        'missed_pace_count' :        0
    },

    /**
     * @description This operation primes the options object for use.
     */
    _prime: function() {
        var options = this.options;
        // update the readouts and range.
        if (options.readout_range_min === null)
            options.readout_range_min = options.range_min;
        if (options.readout_range_max === null)
            options.readout_range_max = options.range_max;

        this.range = options.readout_range_max - options.readout_range_min;

        // normalize the scale of the gauge.
        this.scale_x = options.width / this._normalW;
        this.scale_y = options.height / this._normalH;

        // configure the ticks of the gauge.
        if (options.ticks_major === null)
            options.ticks_major = this.range / 5.0;
        if (options.ticks_minor === null)
            options.ticks_minor = options.ticks_major / 5.0;
    },

    /**
     * @private
     * @description This operation initializes the runtime data and
     *              initiates the asyncronous loading of the gauge GUI.
     */
    _create: function()
    {
        var options = this.options;
        options.channel_readings = [];
        
        // prime the options passed in.
        this._prime();
        
        // initialize the channels
        var channel_number = options.channels;
        while (channel_number--)
        {
            options.channel_readings[channel_number] = [0];
        }
        
        // update the 'resolution' to equal the physical pixel count
        options.resolution = this._lineBase;
        if (options.true_pixels)
        {
            options.resolution = this._lineBase * this.scale_x;
        }

        // update the DOM element for this gauge.
        var self = $(this.element[0]);
        self.addClass('dial_gauge').css({
            // top: options.top + 'px',
            // left: options.left + 'px',
            width: options.width + 'px',
            height: options.height + 'px'
        });
        
        // load in the gauge graphic.
        var handle = this;
        self.svg({loadURL: this._baseURL + 'svg/strip_multi_gauge.svg',
                  onLoad: function(svg, error) {
                          handle._ready(svg, error);} });
    },

    /**
     * @private
     * @description This operation applies any final operations to the DOM
     *              in preparation for runtime use. It expects the return
     *              signature produced by the SVG plug-in callback.
     * @param svg This argument provides access to the svg object.
     * @param error This argument, when defined, indicates an error.
     */
    _ready : function(svg, error)
    {
        var options = this.options;
        if (typeof(error) !== 'undefined')
        {
            //if (typeof(console.log) !== 'undefined')
            //{
            //    console.log(error);
            //}
        }
        else
        {
            var root = svg.root();
            this.svg = svg;

            var channel_number = options.channels;
            var dom_channels = $('.gaugeChannels', root)[0];

            // add the lines
            var line_group = svg.group(dom_channels,
                                       { transform : 'translate(0,31.5)' });
            // prepare common variables
            var stroke_width = 1.0 ;
            if (options.true_pixels)
            {
                stroke_width = 1.0 / this.scale_x;
            }

            // add all of the rendered lines
            while(channel_number--)
            {
                var line_fill = 'none';
                if (options.filled &&
                    options.mode !== 'min_max')
                {
                    line_fill = 'url(#line' + channel_number + '_mat)';
                }

                svg.path(line_group,
                         'M0,0 l146,0',
                         { 'class'  : ('gaugeLine' +
                                       channel_number +
                                       ' line'),
                           'fill'   : line_fill,
                           'stroke-width': stroke_width });
            }
            // add the gauge ticks group
            this.ticks1 = svg.path(dom_channels,
                                   'M0,0 l10,10',
                                   { 'class' : 'gaugeTicks ticks1_' +
                                               options.skin,
                                     'fill'  : 'url(#dark_mat)' });
            this.ticks2 = svg.path(dom_channels,
                                   'M0,0 l10,10',
                                   { 'class' : 'gaugeTicks ticks2_' +
                                               options.skin,
                                     'fill'  : 'url(#dark_mat)' });
            
            // add the needles
            if (options.show_needle &&
                options.mode !== 'min_max')
            {
                channel_number = options.channels;
                while(channel_number--)
                {
                    var group = svg.group(dom_channels,
                                      { 'class' : ('gaugeNeedle' +
                                                   channel_number) });
                    svg.path(group,
                             'M152,33.5 l-8,-2 l8,-2 z',
                             { 'class'  : 'needle',
                               'fill'   : 'url(#needle' +
                                          channel_number +
                                          '_mat)' });
                }
            }

            if (options.skin === 'white')
            {
                $('rect.gaugeMask', root).each(function(index)
                    {
                        this.setAttribute('style', 'fill:url(#light_mat)');
                    }
                );
               $('.gaugeFont', root)[0].setAttribute('fill', '#0c0c0d');
            }

            // render the guage ticks
            this._drawTicks();

            this.ready = true;

            var handle = this;
            options.timer = window.setTimeout(function(){
                handle._update();}, options.pace);

            this._refresh();
        }
    },

    _update : function()
    {
        var options = this.options;

        var minShow = 1;
        if (options.keep_full)
        {
            minShow = options.resolution;
            if (options.mode === 'min_max')
            {
                minShow = minShow * 2;
            }
        }

        var missed_pace = true;
        var channel_number = options.channels;
        while(channel_number--)
        {
            var channel = options.channel_readings[channel_number];
            if (channel.length > minShow)
            {
                if (options.mode === 'linear')
                {
                    channel.splice(0,1);
                }
                else if (options.mode === 'min_max')
                {
                    channel.splice(0,2);
                }
                else
                {
                    //if (typeof(console.log) !== 'undefined')
                    //{
                    //    console.log('Invalid mode');
                        return;
                    //}
                }
                missed_pace = false;
            }
        }
        if (missed_pace)
        {
            options.missed_pace_count++;
        }
        else
        {
            options.missed_pace_count = 0;
        }

        var handle = this;
        options.timer = window.setTimeout(function(){
            handle._update();}, options.pace);

        this._drawLines();
    },

    _drawTicks : function()
    {
        var ticks = '';
        var options = this.options;
        var dom_ticks1 = this.ticks1;
        var dom_ticks2 = this.ticks2;
        var _meterFactor = this._meterFactor;
        var _lineBase = this._lineBase;
        var y,py;
        if (options.ticks_minor)
        {
            for (y = options.readout_range_min;
                 y <= options.readout_range_max;
                 y += options.ticks_minor)
            {
                // provide a modulo with a resolution of 0.000
                var modulo = Math.round((y - options.readout_range_min) * 1000.0) %
                             (options.ticks_major * 1000.0);
                if (modulo)
                {
                    py = (this._percent(y) * _meterFactor) + 0.5;
                    ticks += 'M0,' + py + ' l5,0 M' +
                             _lineBase +',' + py + ' l-5,0 ';
                }
            }
        }
        dom_ticks1.setAttribute('d', ticks);
        if (options.ticks_major)
        {
            for (y = options.readout_range_min;
                 y <= options.readout_range_max;
                 y += options.ticks_major)
            {
                py = (this._percent(y) * _meterFactor) + 0.5;
                ticks += 'M0,' + py + ' l10,0 M' +
                         _lineBase +',' + py + ' l-10,0 ';
            }
        }
        dom_ticks2.setAttribute('d', ticks);
    },

    _drawLines : function()
    {
        var options = this.options;
        var _meterFactor = this._meterFactor;
        var channel_number = options.channels;
        var response = '';
        var mode = options.mode;
        var _lineFactor = this._lineBase / options.resolution;
        if (mode === 'min_max')
        {
            // half step the line factor.
            _lineFactor = _lineFactor / 2.0;
        }
        var x,r,r_min,r_max;
        var resolution = options.resolution;
        var show_needle = options.show_needle;
        var filled = options.filled;

        var svg_root = this.svg.root();

        while (channel_number--)
        {
            var channelLine = $('.gaugeLine' + channel_number, svg_root);

            var reading = options.channel_readings[channel_number];
            var numReadings = reading.length;

            if (mode === 'linear')
            {
                // only show as many readings as there is room to display.
                if (numReadings > resolution)
                {
                    numReadings = resolution;
                }

                // calculate the percent of the needle.
                if (show_needle)
                {
                    var absolute = _meterFactor *
                                   this._percent(reading[numReadings - 1]);
                    $('.gaugeNeedle' + channel_number, svg_root).attr({
                        transform: 'translate(0, -' + absolute + ')'
                    });
                }
                // calculate the line.
                var initial = _meterFactor *
                              this._percent(reading[0]);
                response = 'M';
                if (filled)
                {
                    response += '0,0 L0,-' + _meterFactor + ' L';
                }
                response += '0,' + (-initial);
                for (x = 1; x < numReadings; x++)
                {
                    r = _meterFactor *
                        this._percent(reading[x]);
                    response += ' L' + (x * _lineFactor) + ',' + (-r);
                }

                // close the loop for fill rendering.
                if (options.filled)
                {
                    response += ' L' +
                    ((numReadings - 1) * _lineFactor) + ',0 L0,0 Z';
                }
            }
            else if (mode === 'min_max')
            {
                // only show as many readings as there is room to display.
                if (numReadings > resolution * 2)
                {
                    numReadings = resolution * 2;
                }

                // calculate the line.
                response = '';
                for (x = 0; x < numReadings - 1; x+=2)
                {
                    r_min = _meterFactor *
                            this._percent(reading[x]);
                    r_max = _meterFactor *
                            this._percent(reading[x+1]);
                    // force resolution
                    if (r_min === r_max)
                    {
                        r_max += 0.25;
                    }
                    var _x = x * _lineFactor;
                    response += ' M' + (_x) + ',' + (-r_min) +
                                ' L' + (_x) + ',' + (-r_max);
                }
            }
            $(channelLine, svg_root).attr({d:response});
        }
    },

    /**
     * @private
     * @description This operation takes no arguments. It will update all
     *              GUI elements of the SVG Gauge. It will have no effect
     *              if '_ready' is FALSE.
     */
    _refresh : function()
    {
        if (!this.ready)
            return;
        var options = this.options;
        var scale_x = this.scale_x;
        var scale_y = this.scale_y;
        var lowerScale = Math.min(scale_x, scale_y);

        var svg_root = this.svg.root();

        // update the rendering of the gauge.
        $('.gauge', svg_root).attr(
            'transform', 'scale(' + scale_x + ',' + scale_y + ')'
        );

        $('.gaugeLower', svg_root).text(options.readout_range_min).attr(
            {'x'         : this.lowerX * scale_x,
             'y'         : this.lowerY * scale_y,
             'font-size' : (this.rangeFont * lowerScale) + '%'});

        $('.gaugeUpper', svg_root).text(options.readout_range_max).attr(
            {'x'         : this.upperX * scale_x,
             'y'         : this.upperY * scale_y,
             'font-size' : (this.rangeFont * lowerScale) + '%'});

        $('.gaugeUnit', svg_root).text(options.unit).attr(
            {'x'         : this.unitX * scale_x,
             'y'         : this.unitY * scale_y,
             'font-size' : (this.unitFont * lowerScale) + '%'});

         $('.gaugeLegend', svg_root).text(options.legend).attr(
            {'x'         : this.legendX * scale_x,
             'y'         : this.legendY * scale_y,
             'font-size' : (this.legendFont * lowerScale) + '%'});

        this._drawLines();
    },

    /**
     * @private
     * @description This operation calculates the placement of a reading
     *              as a percentage of the range.
     * @param reading This argument defines the reading to be evaluated.
     * @return This operation returns a value in % where 'min' = '0'
     *         percent and 'max' = '100' percent.
     */
    _percent: function(reading)
    {
        var options = this.options;
        return ((reading - options.readout_range_min) / this.range);
    },

    /**
     * @description This operation provides access to this widgets options
     *              to the outside world.
     */
    get_configuration : function()
    {
        return this.options;
    },

    /**
     * @description This operation updates the current configuration of
     *              the gauge and refreshes the display.
     * @param new_config This argument expects and object containing key
     *                   value pairs to be updated. Unspecified
     *                   configuration items will not be changed.
     */
    set_configuration : function( new_config )
    {
        if ( new_config ) {
            var options = this.options;
            // Merge in the new configuration items.
            $.extend( options, new_config );
            // prime the options passed in.
            this._prime();

            // skip past render updates if not ready.
            if (this.ready && !options.no_svg)
            {
                // update the root to the new size.
                $('.gauge', this.svg.root()).parent().attr({
                    height: options.height,
                    width: options.width
                });
                this._drawTicks();
            }

            // redraw the gauge.
            this._refresh();
        }
    },

    /**
     * @description This operation updates the current reading of the
     *              gauge and updates the GUI when '_ready' is TRUE.
     * @param channel This argument defines the channel being set.
     * @param updated_reading This argument defines the new array FLOAT values
     *                        that this gauge should reflect.
     * NOTE: Repeat channel and updated_reading as many times as needed.
     */
    set_reading : function()
    {
        var options = this.options;
        // array the arguments and step through them until exhausted.
        var args = Array.prototype.slice.call(arguments);
        var channel, updated_reading;

        // if only a single value is submitted, just update the first channel.
        // this is for compatibility with the other readout set_reading
        // functions.
        if (args.length === 1)
        {
            var one_shot = options.channel_readings[0];
            one_shot[one_shot.length] = args[0];
        }

        while(args.length >= 2)
        {
            channel = args[0];
            updated_reading = args[1];
        

            if (channel >= options.channels)
            {
                // if (typeof(console.log) !== 'undefined')
                //{
                //    console.log('Cannot set an invalid channel');
                //}
                return false;
            }

            var reading = options.channel_readings[channel];

            var check = updated_reading.length;

            // peg the needle and append them.
            var min = options.min;
            var max = options.max;
            while (check--)
            {
                var updated = parseFloat(updated_reading[check]);
                if (updated < min)
                {
                    updated = min;
                }
                if (updated > max)
                {
                    updated = max;
                }
                // apply the values
                reading[reading.length] = updated;
            }

            // crop the reading by the missed pace count.
            var spliceSize = options.missed_pace_count;
            if (options.keep_full)
            {
                // ensure that the reading is not over cropped
                var projected = reading.length - spliceSize;
                if (projected < options.resolution)
                {
                    spliceSize = reading.length - options.resolution;
                }
            }
            reading.splice(0, spliceSize);

            // remove the processed args.
            args.splice(0,2);
        }

        // If the internal thread is not already running, start it.
        if (options.timer === null && this.ready)
        {
            // TODO: set this pace in a way to sync it.

            var handle = this;
            options.timer = window.setTimeout(function(){
               handle._update();}, options.pace);
        }
        return this;
    },

    destroy: function() {
        var self = $(this.element[0]);
        self.svg('destroy');

        $.Widget.prototype.destroy.apply(this, arguments);
    }
});

}( jQuery ));

